
def op_clb(
        lnk: float,
        bio: float,
):
    """
    Computes the OP_CLB metric. As described by
        [celltrackingchallenge](http://celltrackingchallenge.net/).

    Args:
        lnk: The linking metric.
        bio: The biological metric.

    Returns:
        The OP_CLB metric.
    """

    op = 0.5 * lnk + 0.5 * bio
    return op



import argparse
from os.path import join, basename, dirname, exists
import copy
from multiprocessing import Pool, cpu_count
import numpy as np
import pandas as pd

from ctc_metrics.metrics import ALL_METRICS
from ctc_metrics.utils.filesystem import parse_directories, read_tracking_file, \
    parse_masks

from ctc_metrics.scripts.evaluate import match_computed_to_reference_masks, \
    calculate_metrics


def load_data(
        gt: str,
        threads: int = 0,
):
    """
    Load the data from the ground truth and use it as computed and reference
    data.

    Args:
        gt: The path to the ground truth.
        threads: The number of threads to use for multiprocessing.

    Returns:
        The computed and reference tracks, the trajectories and the segmentation
        masks.
    """
    # Read tracking files and parse mask files
    ref_tracks = read_tracking_file(join(gt, "TRA", "man_track.txt"))
    comp_tracks = np.copy(ref_tracks)
    ref_tra_masks = parse_masks(join(gt, "TRA"))
    comp_masks = ref_tra_masks
    assert len(ref_tra_masks) > 0, f"{gt}: Ground truth masks is 0!)"
    # Match golden truth tracking masks to result masks
    traj = match_computed_to_reference_masks(
        ref_tra_masks, comp_masks, threads=threads)
    # Match golden truth segmentation masks to result masks
    segm = {}
    return comp_tracks, ref_tracks, traj, segm, comp_masks


def remove_mitosis(
        comp_tracks: np.ndarray,
        num_to_remove: int,
        seed: int,
):
    """
    Remove mitosis events by removing the mother daughter relation.

    Args:
        comp_tracks: The computed tracks.
        num_to_remove: The number of mitosis events to remove.
        seed: The seed for the random number generator.

    Returns:
        The computed tracks with the mitosis events removed.
    """
    if num_to_remove == 0:
        return comp_tracks
    parents, counts = np.unique(
        comp_tracks[comp_tracks[:, 3] > 0, 3], return_counts=True)
    parents = parents[counts > 1]
    num_splits = min(num_to_remove, len(parents))
    np.random.seed(seed)
    np.random.shuffle(parents)
    for parent in parents[:num_splits]:
        comp_tracks[np.isin(comp_tracks[:, 3], parent), 3] = 0
    return comp_tracks


def sample_fn(l_comp, max_num_candidates, seed):
    """
    Sample false negatives.
    """
    candidates = []
    for frame, x in enumerate(l_comp):
        for i, _ in enumerate(x):
            candidates.append((frame, i))
    np.random.seed(seed)
    np.random.shuffle(candidates)
    num_fn = min(max_num_candidates, len(candidates))
    return candidates[:num_fn]


def add_false_negatives(
        comp_tracks: np.ndarray,
        traj: dict,
        noise_add_false_negative: int,
        seed: int
):
    """
    Add false negatives to the data.

    Args:
        comp_tracks: The computed tracks.
        traj: The trajectories.
        noise_add_false_negative: The number of false negatives to add.
        seed: The seed for the random number generator.

    Returns:
        The computed tracks and the trajectories with the false negatives added.
    """
    if noise_add_false_negative == 0:
        return comp_tracks, traj

    next_id = np.max(comp_tracks[:, 0]) + 1
    l_comp = traj["labels_comp"]
    m_comp = traj["mapped_comp"]
    m_ref = traj["mapped_ref"]
    for frame, i in sample_fn(l_comp, noise_add_false_negative, seed):
        if i >= len(l_comp[frame]):
            i = np.random.randint(len(l_comp[frame]))
        v = l_comp[frame][i]
        # Remove from current frame
        while v in m_comp[frame]:
            _i = m_comp[frame].index(v)
            m_comp[frame].pop(_i)
            m_ref[frame].pop(_i)
        l_comp[frame].pop(i)
        # Create new trajectory
        start, end = comp_tracks[comp_tracks[:, 0] == v, 1:3][0]
        if start == end:
            comp_tracks = comp_tracks[comp_tracks[:, 0] != v]
            comp_tracks[comp_tracks[:, 3] == v, 3] = 0
        elif frame == start:
            comp_tracks[comp_tracks[:, 0] == v, 1] += 1
        elif frame == end:
            comp_tracks[comp_tracks[:, 0] == v, 2] -= 1
        else:
            comp_tracks[comp_tracks[:, 0] == v, 2] = frame - 1
            comp_tracks[comp_tracks[:, 3] == v, 3] = next_id
            comp_tracks = np.concatenate(
                [comp_tracks, [[next_id, frame + 1, end, v]]], axis=0)
            for f in range(frame + 1, end + 1):
                _l_comp = np.asarray(l_comp[f])
                _m_comp = np.asarray(m_comp[f])
                _l_comp[_l_comp == v] = next_id
                _m_comp[_m_comp == v] = next_id
                l_comp[f] = _l_comp.tolist()
                m_comp[f] = _m_comp.tolist()
        next_id += 1
    return comp_tracks, traj


def add_false_positives(
        comp_tracks: np.ndarray,
        traj: dict,
        noise_add_false_positive: int,
        seed: int,
):
    """
    Add false positives to the data.

    Args:
        comp_tracks: The computed tracks.
        traj: The trajectories.
        noise_add_false_positive: The number of false positives to add.
        seed: The seed for the random number generator.

    Returns:
        The computed tracks and the trajectories with the false positives added.
    """
    if noise_add_false_positive == 0:
        return comp_tracks, traj

    label = traj["labels_comp"]
    next_id = np.max(comp_tracks[:, 0]) + 1
    max_frame = np.max(comp_tracks[:, 2])
    fp_to_add = int(noise_add_false_positive)
    np.random.seed(seed)
    for _ in range(fp_to_add):
        frame = np.random.randint(max_frame + 1)
        comp_tracks = np.concatenate(
            [comp_tracks, [[next_id, frame, frame, 0]]], axis=0)
        label[frame].append(next_id)
        next_id += 1

    return comp_tracks, traj


def remove_matches(
        comp_tracks: np.ndarray,
        traj: dict,
        noise_remove_matches: int,
        seed: int,
):
    """
    Remove ref-comp matches from the data.

    Args:
        comp_tracks: The computed tracks.
        traj: The trajectories.
        noise_remove_matches: The number of matches to remove.
        seed: The seed for the random number generator.

    Returns:
        The computed tracks and the trajectories with the matches removed.
    """
    if noise_remove_matches == 0:
        return comp_tracks, traj

    m_comp = traj["mapped_comp"]
    m_ref = traj["mapped_ref"]
    candidates = []
    for frame in range(1, len(m_comp)):
        for i in range(len(m_comp[frame])):
            candidates.append(frame)
    np.random.seed(seed)
    np.random.shuffle(candidates)
    num_unassoc = min(noise_remove_matches, len(candidates))
    for frame in candidates[:num_unassoc]:
        total_inds = len(m_comp[frame])
        i = np.random.randint(total_inds)
        m_comp[frame].pop(i)
        m_ref[frame].pop(i)

    return comp_tracks, traj


def add_id_switches(
        comp_tracks: np.ndarray,
        traj: dict,
        noise_add_idsw: int,
        seed: int,
):
    """
    Add ID switches to the data.

    Args:
        comp_tracks: The computed tracks.
        traj: The trajectories.
        noise_add_idsw: The number of ID switches to add.
        seed: The seed for the random number generator.

    Returns:
        The computed tracks and the trajectories with the ID switches added.
    """
    if noise_add_idsw == 0:
        return comp_tracks, traj

    labels_comp = traj["labels_comp"]
    m_comp = traj["mapped_comp"]
    candidates = []
    for frame, x in enumerate(m_comp):
        if np.unique(x).shape[0] <= 1:
            continue
        for _ in range(len(np.unique(x)) - 1):
            candidates.append(frame)
    np.random.seed(seed)
    np.random.shuffle(candidates)
    num_unassoc = min(noise_add_idsw, len(candidates))
    for frame in candidates[:num_unassoc]:
        # Select two random indices
        comp = m_comp[frame]
        c1, c2 = np.random.choice(comp, 2, replace=False)
        end1 = int(comp_tracks[comp_tracks[:, 0] == c1, 2].squeeze())
        end2 = int(comp_tracks[comp_tracks[:, 0] == c2, 2].squeeze())
        children1 = comp_tracks[:, 3] == c1
        children2 = comp_tracks[:, 3] == c2
        # Swap the two indices
        for f in range(frame, max(end1, end2) + 1):
            _l_comp = np.asarray(labels_comp[f])
            _comp = np.asarray(m_comp[f])
            i1 = _comp == c1
            i2 = _comp == c2
            _comp[i1] = c2
            _comp[i2] = c1
            i1 = _l_comp == c1
            i2 = _l_comp == c2
            _l_comp[i1] = c2
            _l_comp[i2] = c1
            labels_comp[f] = _l_comp.tolist()
            m_comp[f] = _comp.tolist()
        i1 = comp_tracks[:, 0] == c1
        i2 = comp_tracks[:, 0] == c2
        comp_tracks[i1, 2] = end2
        comp_tracks[i2, 2] = end1
        comp_tracks[children1, 3] = c2
        comp_tracks[children2, 3] = c1

    return comp_tracks, traj


def add_noise(
        comp_tracks: np.ndarray,
        traj: dict,
        seed: int = 0,
        noise_add_false_negative: int = 0,
        noise_add_false_positive: int = 0,
        noise_add_idsw: int = 0,
        noise_remove_matches: int = 0,
        noise_remove_mitosis: int = 0,
):
    """
    Add noise to the data.

    Args:
        comp_tracks: The computed tracks.
        traj: The trajectories.
        seed: The seed for the random number generator.
        noise_add_false_negative:
            Adds n false negatives to the data, where the parameter is n.
        noise_remove_mitosis:
            Removes parend daughter relations of n mitosis events, where the
            parameter describes n.
        noise_add_false_positive:
            Adds n false positives to the data, where the parameter is n.
        noise_add_idsw:
            Adds n ID switches to the data, where the parameter is n.
        noise_remove_matches:
            Removes n matches from the data, where the parameter is n.
            This produces n false negatives and n false positives.

    Returns:
        comp_tracks: Updated with noise applied.
        traj: Updated with noise applied.
    """
    comp_tracks = np.copy(comp_tracks)
    traj = copy.deepcopy(traj)

    # Remove children of mitosis events
    comp_tracks = remove_mitosis(
        comp_tracks, noise_remove_mitosis, seed)

    # Add false negatives
    comp_tracks, traj = add_false_negatives(
        comp_tracks, traj, noise_add_false_negative, seed)

    # Add false positives
    comp_tracks, traj = add_false_positives(
        comp_tracks, traj, noise_add_false_positive, seed)

    # Unmatch true positives
    comp_tracks, traj = remove_matches(
        comp_tracks, traj, noise_remove_matches, seed)

    # Add IDSw
    comp_tracks, traj = add_id_switches(
        comp_tracks, traj, noise_add_idsw, seed)

    return comp_tracks, traj


def is_new_setting(
        setting: dict,
        path: str,
        name: str,
        df: pd.DataFrame = None,
):
    """
    Check if the setting parameter setting is already existing in the csv file.

    Args:
        setting: The setting to check.
        path: The path to the csv file.
        name: The name of the sequence.
        df: The dataframe to check.

    Returns:
        True if the setting is new, False otherwise.
    """
    if exists(path):
        setting["name"] = name
        if df is None:
            df = pd.read_csv(path, index_col="index", sep=";")
        _df = df.copy()
        for k, v in setting.items():
            _df = _df[_df[k] == v]
            if len(_df) == 0:
                return True, df
        return False, df
    return True, df


def append_results(
        path: str,
        results: list,
):
    """
    Append the results to the csv file.

    Args:
        path: The path to the csv file.
        results: The results to append.
    """
    # Check if the file exists
    results = [pd.DataFrame.from_dict(r, orient="index").T for r in results]
    if exists(path):
        df = pd.read_csv(path, index_col="index", sep=";")
        df = pd.concat([df] + results)
        df.reset_index(drop=True, inplace=True)
    else:
        df = pd.concat(results)
    df.to_csv(path, index_label="index", sep=";")


def run_noisy_sample(
        comp_tracks: np.ndarray,
        ref_tracks: np.ndarray,
        traj: dict,
        segm: dict,
        metrics: list,
        name: str,
        setting: dict,
        default_setting: dict,
):
    """
    Run a noisy sample

    Args:
        comp_tracks: The computed tracks.
        ref_tracks: The reference tracks.
        traj: The trajectories.
        segm: The segmentation masks.
        metrics: The metrics to calculate.
        name: The name of the sequence.
        setting: The noise setting.
        default_setting: The default setting without noise.

    Returns:
        The results stored in a dictionary.
    """
    # Add noise to the data and calculate the metrics
    n_comp_tracks, n_traj = add_noise(
        comp_tracks, traj, **setting)

    results = {"name": name}
    results.update(default_setting)

    resulting_metrics = calculate_metrics(
        n_comp_tracks, ref_tracks, n_traj, segm, metrics,
        is_valid=True
    )
    results.update(resulting_metrics)
    return results


def filter_existing_noise_settings(
        noise_settings: list,
        csv_file: str,
        name: str,
):
    """
    Filter and remove existing noise settings from the list of noise settings.

    Args:
        noise_settings: The list of noise settings.
        csv_file: The path to the csv file.
        name: The name of the sequence.

    Returns:
        The list of new noise settings.
    """
    df = None
    new_noise_settings = []

    for _, setting in enumerate(noise_settings):
        # Check if noise setting is new
        default_setting = {
            "seed": 0,
            "noise_add_false_positive": 0,
            "noise_add_false_negative": 0,
            "noise_add_idsw": 0,
            "noise_remove_matches": 0,
            "noise_remove_mitosis": 0,
        }

        default_setting.update(setting)

        is_new, df = is_new_setting(default_setting, csv_file, name, df)
        if not is_new:
            continue
        new_noise_settings.append((setting, default_setting))
    return new_noise_settings


def create_noise_settings(
        repeats: int,
        num_false_neg: int,
        num_false_pos: int,
        num_idsw: int,
        num_matches: int,
        comp_tracks: np.ndarray,
        ref_tracks: np.ndarray,
):
    """
    Create a list of noise settings that should be executed from the given
    parameters.

    Args:
        repeats: The number of repeats for each noise setting.
        num_false_neg: The number of false negatives to add.
        num_false_pos: The number of false positives to add.
        num_idsw: The number of ID switches to add.
        num_matches: The number of matches to remove.
        comp_tracks: The computed tracks.
        ref_tracks: The reference tracks.

    Returns:
        The list of noise settings.
    """
    # Extract some statistics
    parents, counts = np.unique(
        comp_tracks[comp_tracks[:, 3] > 0, 3], return_counts=True)
    num_false_negs_max = int(np.sum(ref_tracks[:, 2] - ref_tracks[:, 1] + 1))

    # Create dictionary
    noise_settings = [{}]

    for i in range(0, repeats):
        # Add mitosis detection noise
        for x in range(1, len(parents[counts > 1]) + 1):
            noise_settings.append({"seed": i, "noise_remove_mitosis": x})

        # Add false negative noise
        for x in range(1, min(num_false_neg, num_false_negs_max)):
            noise_settings.append({"seed": i, "noise_add_false_negative": x})

        # Add false positive noise
        for x in range(1, num_false_pos):
            noise_settings.append({"seed": i, "noise_add_false_positive": x})

        # Add matching noise
        for x in range(1, min(num_matches, num_false_negs_max)):
            noise_settings.append({"seed": i, "noise_remove_matches": x})

        # Add ID switch noise
        for x in range(1, num_idsw):
            noise_settings.append({"seed": i, "noise_add_idsw": x})

    return noise_settings


def evaluate_sequence(
        gt: str,
        name: str,
        threads: int = 0,
        csv_file: str = None,
        save_after: int = 20,
        repeats: int = 10,
        num_false_neg: int = 500,
        num_false_pos: int = 500,
        num_idsw: int = 500,
        num_matches: int = 500,
):
    """
    Evaluates a single sequence

    Args:
        gt: The path to the ground truth.
        name: The name of the sequence.
        threads: The number of threads to use for multiprocessing.
        csv_file: The path to the csv file to store the results.
        save_after: Save results after n runs.
        repeats: The number of repeats for each noise setting.
        num_false_neg: The number of false negatives to add.
        num_false_pos: The number of false positives to add.
        num_idsw: The number of ID switches to add.
        num_matches: The number of matches to remove.

    """

    print("Run noise test on ", gt, end="...")
    # Prepare all metrics
    metrics = copy.deepcopy(ALL_METRICS)
    metrics.remove("Valid")
    metrics.remove("SEG")

    comp_tracks, ref_tracks, traj, segm, _ = load_data(gt, threads)

    # Selection of noise settings
    noise_settings = create_noise_settings(
        repeats, num_false_neg, num_false_pos, num_idsw, num_matches,
        comp_tracks, ref_tracks
    )

    # Filter existing noise settings
    new_noise_settings = filter_existing_noise_settings(
        noise_settings, csv_file, name
    )

    # Evaluate new noise settings
    if threads == 1:
        results_list = []
        for i, (setting, default_setting) in enumerate(new_noise_settings):
            print(
                f"\rRun noise test on {gt}, \t{i + 1}\t/ {len(new_noise_settings)}",
                end=""
            )
            # Add noise to the data and calculate the metrics
            results = run_noisy_sample(
                comp_tracks, ref_tracks, traj, segm, metrics,
                name, setting, default_setting
            )
            # Aggregate results and store them every n runs
            results_list.append(results)
            if len(results_list) == save_after or i + 1 == len(new_noise_settings):
                append_results(csv_file, results_list)
                results_list = []

    else:
        threads = cpu_count() if threads == 0 else threads
        with Pool(threads) as p:
            input_list = []
            for i, (setting, default_setting) in enumerate(new_noise_settings):
                print(
                    f"\rRun noise test on {gt}, \t{i + 1}\t/ {len(new_noise_settings)}",
                    end=""
                )
                # Add noise to the data and calculate the metrics
                input_list.append((
                    comp_tracks, ref_tracks, traj, segm, metrics,
                    name, setting, default_setting
                ))
                # Process in parallel and
                if len(input_list) == save_after or i + 1 == len(new_noise_settings):
                    results_list = p.starmap(run_noisy_sample, input_list)
                    append_results(csv_file, results_list)
                    input_list = []
    print("")


def evaluate_all(
        gt_root: str,
        csv_file: str = None,
        threads: int = 0,
        **kwargs
):
    """
    Evaluate all sequences in a directory

    Args:
        gt_root: The root directory of the ground truth.
        csv_file: The path to the csv file to store the results.
        threads: The number of threads to use for multiprocessing.
        **kwargs: The noise settings.

    """
    ret = parse_directories(gt_root, gt_root)
    for _, gt, name in zip(*ret):
        evaluate_sequence(gt, name, threads, csv_file, **kwargs)


def parse_args():
    """ Parse arguments """
    parser = argparse.ArgumentParser(description='Evaluates CTC-Sequences.')
    parser.add_argument('--gt', type=str, required=True)
    parser.add_argument('-r', '--recursive', action="store_true")
    parser.add_argument('--csv-file', type=str, default=None)
    parser.add_argument('-n', '--num-threads', type=int, default=0)
    parser.add_argument('--num-false-pos', type=int, default=500)
    parser.add_argument('--num-false-neg', type=int, default=500)
    parser.add_argument('--num-idsw', type=int, default=500)
    parser.add_argument('--num-matches', type=int, default=500)
    parser.add_argument('--save-after', type=int, default=100)
    parser.add_argument('--repeats', type=int, default=10)
    args = parser.parse_args()
    return args


def main():
    """
    Main function that is called when the script is executed.
    """
    args = parse_args()

    # Evaluate sequence or whole directory
    experiments = {
        "num_false_pos": args.num_false_pos,
        "num_false_neg": args.num_false_neg,
        "num_idsw": args.num_idsw,
        "num_matches": args.num_matches,
        "repeats": args.repeats,
        "save_after": args.save_after,
    }
    if args.recursive:
        evaluate_all(args.gt, args.csv_file, args.num_threads, **experiments)
    else:
        challenge = basename(dirname(args.gt))
        sequence = basename(args.gt).replace("_GT", "")
        name = challenge + "_" + sequence
        evaluate_sequence(
            args.gt, name, args.num_threads, args.csv_file, **experiments
        )


if __name__ == "__main__":
    main()




def det(
        NS: int,
        FN: int,
        FP: int,
        num_vertices: int,
        **_ignored
):
    """
    Calculate Detection (DET) metric.

    According to
        Cell Tracking Accuracy Measurement Based on Comparison of Acyclic
        Oriented Graphs; Matula etal. 2015

    Args:
        NS: Split vertex operations.
        FN: Number of false negatives.
        FP: Number of false positives.
        num_vertices: Number of vertices in the graph.
        **_ignored: Ignored arguments.

    Returns:
        The Detection metric.

    """
    # Calculate AOGM_D
    w_ns = 5
    w_fn = 10
    w_fp = 1
    AOGM_D = w_ns * NS + w_fn * FN + w_fp * FP
    # Calculate AOGM_D0 (create graph from scratch)
    AOGM_D0 = w_fn * num_vertices  # All false negatives
    # Calculate DET
    DET = 1 - min(AOGM_D, AOGM_D0) / AOGM_D0
    return float(DET)




from test.utils import test_seq_res, test_seq_gt, test_det, test_seg, \
    test_tra, test_ct, test_tf, test_bc0, test_bc1, test_bc2, test_bc3, test_cca
from ctc_metrics import evaluate_sequence


def test_metric_det():
    max_d = 0.00001
    metrics = evaluate_sequence(
        test_seq_res, test_seq_gt, metrics=["DET"]
    )
    assert abs(metrics["DET"] - test_det) < max_d, \
        f"{metrics['DET']} != {test_det}"


def test_metric_seg():
    max_d = 0.00001
    metrics = evaluate_sequence(
        test_seq_res, test_seq_gt, metrics=["SEG"]
    )
    assert abs(metrics["SEG"] - test_seg) < max_d, \
        f"{metrics['SEG']} != {test_seg}"


def test_metric_tra():
    max_d = 0.00001
    metrics = evaluate_sequence(
        test_seq_res, test_seq_gt, metrics=["TRA"]
    )
    assert abs(metrics["TRA"] - test_tra) < max_d, \
        f"{metrics['TRA']} != {test_tra}"


def test_metric_ct():
    max_d = 0.00001
    metrics = evaluate_sequence(
        test_seq_res, test_seq_gt, metrics=["CT"]
    )
    assert abs(metrics["CT"] - test_ct) < max_d, \
        f"{metrics['CT']} != {test_ct}"


def test_metric_tf():
    max_d = 0.00001
    metrics = evaluate_sequence(
        test_seq_res, test_seq_gt, metrics=["TF"]
    )
    assert abs(metrics["TF"] - test_tf) < max_d, \
        f"{metrics['TF']} != {test_tf}"


def test_metric_bc():
    max_d = 0.00001
    metrics = evaluate_sequence(
        test_seq_res, test_seq_gt, metrics=["BC"]
    )
    assert abs(metrics["BC(0)"] - test_bc0) < max_d, \
        f"{metrics['BC(0)']} != {test_bc0}"
    assert abs(metrics["BC(1)"] - test_bc1) < max_d, \
        f"{metrics['BC(1)']} != {test_bc1}"
    assert abs(metrics["BC(2)"] - test_bc2) < max_d, \
        f"{metrics['BC(2)']} != {test_bc2}"
    assert abs(metrics["BC(3)"] - test_bc3) < max_d, \
        f"{metrics['BC(3)']} != {test_bc3}"


def test_metric_cca():
    metrics = evaluate_sequence(
        test_seq_res, test_seq_gt, metrics=["CCA"]
    )
    assert abs(metrics["CCA"] - test_cca) < 0.00001, \
        f"{metrics['CCA']} != {test_cca}"





def op_ctb(
        seg: float,
        tra: float,
):
    """
    Computes the OP_CTB metric. As described by
        [celltrackingchallenge](http://celltrackingchallenge.net/).

    Args:
        seg: The segmentation metric.
        tra: The tracking metric.

    Returns:
        The OP_CTB metric.
    """

    op = 0.5 * seg + 0.5 * tra
    return op



import argparse
from os import listdir, makedirs
from os.path import join
import tifffile as tiff
import cv2
import numpy as np

from ctc_metrics.utils.filesystem import read_tracking_file


SHOW_BORDER = True
BORDER_WIDTH = {
    "BF-C2DL-HSC": 25,
    "BF-C2DL-MuSC": 25,
    "Fluo-N2DL-HeLa": 25,
    "PhC-C2DL-PSC": 25,
    "Fluo-N2DH-SIM+": 0,
    "DIC-C2DH-HeLa": 50,
    "Fluo-C2DL-Huh7": 50,
    "Fluo-C2DL-MSC": 50,
    "Fluo-N2DH-GOWT1": 50,
    "PhC-C2DH-U373": 50,
}

np.random.seed(0)
PALETTE = np.random.randint(0, 256, (100000, 3))


def get_palette_color(i):
    i = i % PALETTE.shape[0]
    return PALETTE[i]


def visualize(
        img_dir: str,
        res_dir: str,
        viz_dir: str = None,
        video_name: str = None,
        border_width: str = None,
        show_labels: bool = True,
        show_parents: bool = True,
        ids_to_show: list = None,
        start_frame: int = 0,
        framerate: int = 30,
        opacity: float = 0.5,
):  # pylint: disable=too-many-arguments,too-complex
    """
    Visualizes the tracking results.

    Args:
        img_dir: str
            The path to the images.
        res_dir: str
            The path to the results.
        viz_dir: str
            The path to save the visualizations.
        video_name: str
            The path to the video if a video should be created. Note that no
            visualization is available during video creation.
        border_width: str or int
            The width of the border. Either an integer or a string that
            describes the challenge name.
        show_labels: bool
            Print instance labels to the output.
        show_parents: bool
            Print parent labels to the output.
        ids_to_show: list
            The IDs of the instances to show. All others will be ignored.
        start_frame: int
            The frame to start the visualization.
        framerate: int
            The framerate of the video.
        opacity: float
            The opacity of the instance colors.

    """
    # Define initial video parameters
    wait_time = max(1, round(1000 / framerate))
    if border_width is None:
        border_width = 0
    elif isinstance(border_width, str):
        try:
            border_width = int(border_width)
        except ValueError as exc:
            if border_width in BORDER_WIDTH:
                border_width = BORDER_WIDTH[border_width]
            else:
                raise ValueError(
                    f"Border width '{border_width}' not recognized. "
                    f"Existing datasets: {BORDER_WIDTH.keys()}"
                ) from exc

    # Load image and tracking data
    images = [x for x in sorted(listdir(img_dir)) if x.endswith(".tif")]
    results = [x for x in sorted(listdir(res_dir)) if x.endswith(".tif")]
    parents = {
        l[0]: l[3] for l in read_tracking_file(join(res_dir, "res_track.txt"))
    }

    # Create visualization directory
    if viz_dir:
        makedirs(viz_dir, exist_ok=True)

    video_writer = None

    # Loop through all images
    while start_frame < len(images):
        # Read image file
        img_name, res_name = images[start_frame], results[start_frame]
        img_path, res_path,  = join(img_dir, img_name), join(res_dir, res_name)
        print(f"\rFrame {img_name} (of {len(images)})", end="")

        # Visualize the image
        viz = create_colored_image(
            cv2.imread(img_path),
            tiff.imread(res_path),
            labels=show_labels,
            frame=start_frame,
            parents=parents if show_parents else None,
            ids_to_show=ids_to_show,
            opacity=opacity,
        )
        if border_width > 0:
            viz = cv2.rectangle(
                viz,
                (border_width, border_width),
                (viz.shape[1] - border_width, viz.shape[0] - border_width),
                (0, 0, 255), 1
            )

        # Save the visualization
        if video_name is not None:
            if video_writer is None:
                video_path = join(
                    viz_dir, f"{video_name.replace('.mp4', '')}.mp4")
                video_writer = cv2.VideoWriter(
                    video_path,
                    cv2.VideoWriter_fourcc(*"mp4v"),
                    framerate,
                    (viz.shape[1], viz.shape[0])
                )
            video_writer.write(viz)
            start_frame += 1
            continue

        # Show the video
        cv2.imshow("VIZ", viz)
        key = cv2.waitKey(wait_time)
        if key == ord("q"):
            # Quit the visualization
            break
        if key == ord("w"):
            # Start or stop the auto visualization
            if wait_time == 0:
                wait_time = max(1, round(1000 / framerate))
            else:
                wait_time = 0
        elif key == ord("d"):
            # Move to the next frame
            start_frame += 1
            wait_time = 0
        elif key == ord("a"):
            # Move to the previous frame
            start_frame -= 1
            wait_time = 0
        elif key == ord("l"):
            # Toggle the show labels option
            show_labels = not show_labels
        elif key == ord("p"):
            # Toggle the show parents option
            show_parents = not show_parents
        elif key == ord("s"):
            # Save the visualization
            if viz_dir is None:
                print("Please define the '--viz' argument to save the "
                      "visualizations.")
                continue
            viz_path = join(viz_dir, img_name) + ".jpg"
            cv2.imwrite(viz_path, viz)
        else:
            # Move to the next frame
            start_frame += 1


def create_colored_image(
        img: np.ndarray,
        res: np.ndarray,
        labels: bool = False,
        opacity: float = 0.5,
        ids_to_show: list = None,
        frame: int = None,
        parents: dict = None,
):
    """
    Creates a colored image from the input image and the results.

    Args:
        img: np.ndarray
            The input image.
        res: np.ndarray
            The results.
        labels: bool
            Print instance labels to the output.
        opacity: float
            The opacity of the instance colors.
        ids_to_show: list
            The IDs of the instances to show. All others will be ignored.
        frame: int
            The frame number.
        parents: dict
            The parent dictionary.

    Returns:
        The colored image.
    """
    img = np.clip(img, 0, 255).astype(np.uint8)
    kernel = np.ones((3, 3), dtype=np.uint8)
    for i in np.unique(res):
        if i == 0:
            continue
        if ids_to_show is not None:
            if i not in ids_to_show:
                continue
        mask = res == i
        contour = (mask * 255).astype(np.uint8) - \
                  cv2.erode((mask * 255).astype(np.uint8), kernel)
        contour = contour != 0
        img[mask] = (
            np.round((1 - opacity) * img[mask] + opacity * get_palette_color(i))
        )
        img[contour] = get_palette_color(i)
        if frame is not None:
            cv2.putText(img, str(frame), (10, 30),
                        cv2.FONT_HERSHEY_SIMPLEX, 1, (255, 255, 255), 2)
        if labels:
            # Print label to the center of the object
            y, x = np.where(mask)
            y, x = np.mean(y), np.mean(x)
            text = str(i)
            if parents is not None:
                if i in parents:
                    if parents[i] != 0:
                        text += f"({parents[i]})"
            cv2.putText(img, text, (int(x), int(y)),
                        cv2.FONT_HERSHEY_SIMPLEX, 0.5, (255, 255, 255), 1)
    return img


def parse_args():
    """ Parses the arguments. """
    parser = argparse.ArgumentParser(description='Validates CTC-Sequences.')
    parser.add_argument(
        '--img', type=str, required=True,
        help='The path to the images.'
    )
    parser.add_argument(
        '--res', type=str, required=True, help='The path to the results.'
    )
    parser.add_argument(
        '--viz', type=str, default=None,
        help='The path to save the visualizations.'
    )
    parser.add_argument(
        '--video-name', type=str, default=None,
        help='The path to the video if a video should be created. Note that no '
             'visualization is available during video creation.'
    )
    parser.add_argument(
        '--border-width', type=str, default=None,
        help='The width of the border. Either an integer or a string that '
             'describes the challenge name.'
    )
    parser.add_argument(
        '--show-no-labels', action="store_false",
        help='Print no instance labels to the output.'
    )
    parser.add_argument(
        '--show-no-parents', action="store_false",
        help='Print no parent labels to the output.'
    )
    parser.add_argument(
        '--ids-to-show', type=int, nargs='+', default=None,
        help='The IDs of the instances to show. All others will be ignored.'
    )
    parser.add_argument(
        '--start-frame', type=int, default=0,
        help='The frame to start the visualization.'
    )
    parser.add_argument(
        '--framerate', type=int, default=10,
        help='The framerate of the video.'
    )
    parser.add_argument(
        '--opacity', type=float, default=0.5,
        help='The opacity of the instance colors.'
    )
    args = parser.parse_args()
    return args


def main():
    """
    Main function that is called when the script is executed.
    """
    args = parse_args()
    visualize(
        args.img,
        args.res,
        viz_dir=args.viz,
        video_name=args.video_name,
        border_width=args.border_width,
        show_labels=args.show_no_labels,
        show_parents=args.show_no_parents,
        ids_to_show=args.ids_to_show,
        start_frame=args.start_frame,
        framerate=args.framerate,
        opacity=args.opacity,
    )


if __name__ == "__main__":
    main()



from ctc_metrics.metrics.validation.valid import valid
from ctc_metrics.metrics.biological.bc import bc
from ctc_metrics.metrics.biological.ct import ct
from ctc_metrics.metrics.biological.cca import cca
from ctc_metrics.metrics.biological.tf import tf
from ctc_metrics.metrics.technical.seg import seg
from ctc_metrics.metrics.technical.tra import tra
from ctc_metrics.metrics.technical.det import det
from ctc_metrics.metrics.clearmot.mota import mota
from ctc_metrics.metrics.hota.hota import hota
from ctc_metrics.metrics.hota.chota import chota
from ctc_metrics.metrics.identity_metrics.idf1 import idf1
from ctc_metrics.metrics.others.mt_ml import mtml
from ctc_metrics.metrics.others.faf import faf
from ctc_metrics.metrics.technical.op_ctb import op_ctb
from ctc_metrics.metrics.technical.op_csb import op_csb
from ctc_metrics.metrics.biological.bio import bio
from ctc_metrics.metrics.biological.op_clb import op_clb
from ctc_metrics.metrics.technical.lnk import lnk

ALL_METRICS = [
    "Valid", "CHOTA", "BC", "CT", "CCA", "TF", "SEG", "TRA", "DET", "MOTA",
    "HOTA", "IDF1", "MTML", "FAF", "LNK", "OP_CTB", "OP_CSB", "BIO", "OP_CLB"
]





def print_results(results: dict):
    """
    Prints the results in a nice table.

    Args:
        results: A dictionary containing the results.
    """

    def print_line(metrics: dict):
        """
        Prints a line of the table.

        Args:
            metrics: A list containing the arguments for the line.
        """

        print(*[f"{k}: {'N/A' if v is None else float(v):.5},\t" for k, v
                in metrics.items()])

    if isinstance(results, dict):
        print_line(results)
    elif isinstance(results, list):
        for res in results:
            print(res[0], end=":\t\t")
            print_line(res[1])


def store_results(
        path: str,
        results: dict,
        delimiter: str = ";"):
    """
    Stores the results in a csv file.

    Args:
        path: The path to the csv file.
        results: A dictionary containing the results.
        delimiter: The delimiter for the csv file.
    """
    if not path.endswith(".csv"):
        path += ".csv"
    if isinstance(results, dict):
        keys = results.keys()
        with open(path, "w+", encoding="utf-8") as f:
            f.write(delimiter.join(keys) + "\n")
            f.write(delimiter.join([str(v) for v in results.values()]) + "\n")
    elif isinstance(results, list):
        keys = results[0][1].keys()
        with open(path, "w+", encoding="utf-8") as f:
            f.write('dataset'+delimiter+delimiter.join(keys) + "\n")
            for dataset, res in results:
                f.write(dataset+delimiter)
                f.write(delimiter.join([str(v) for v in res.values()]) + "\n")



import numpy as np

from ctc_metrics.utils.representations import assign_comp_to_ref


def calculate_fractions_fo_computed_tracks(
        ref_tracks: np.ndarray,
        labels_ref: list,
        mapped_ref: list,
        mapped_comp: list,
):
    """
    Returns a dictionary with the fractions of overlap for each computed track.
    Each entry in the dictionary is a dictionary with the computed track id as
    key and a new dictionary as value. In the sub-dictionary, the reference
    track id is the key and the fraction of overlap is the value.

    Args:
        ref_tracks: The ground truth tracks.
        labels_ref: The labels of the ground truth masks.
        mapped_ref: The matched labels of the ground truth masks.
        mapped_comp: The matched labels of the result masks.

    Returns:
        A dictionary with the fractions of overlap for each computed track.
    """
    track_assignments = assign_comp_to_ref(labels_ref, mapped_ref, mapped_comp)
    comp_fractions = {}
    for k, v in track_assignments.items():
        start_ref = ref_tracks[ref_tracks[:, 0] == k][0, 1]
        end_ref = ref_tracks[ref_tracks[:, 0] == k][0, 2]
        length = int(end_ref - start_ref + 1)
        array = v[start_ref:end_ref + 1]
        assigned_labels = np.unique(array)
        if len(assigned_labels) == 1 and assigned_labels[0] == 0:
            continue
        assignments = 0
        last_i = 0
        for i in array:
            if last_i != i or i == 0:
                assignments = 0
            if i > 0:
                assignments += 1
                if i not in comp_fractions:
                    comp_fractions[i] = {}
                comp_fractions[i][k] = max(
                    comp_fractions[i].get(k, 0), assignments / length)
            last_i = i
    return comp_fractions


def tf(
        ref_tracks: np.ndarray,
        labels_ref: list,
        mapped_ref: list,
        mapped_comp: list
):
    """
    Computes the track fractions metric. As described in the paper,
         "An objective comparison of cell-tracking algorithms."
           - Vladimir Ulman et al., Nature methods 2017

    Args:
        ref_tracks: The ground truth tracks. A (n,4) numpy ndarray with columns:
            - label
            - birth frame
            - end frame
            - parent
        labels_ref: The labels of the ground truth masks. A list of length
        mapped_ref: The matched labels of the ground truth masks. A list of
            length equal to the number of frames. Each element is a list with
            the matched labels of the ground truth masks in the respective
            frame. The elements are in the same order as the corresponding
            elements in mapped_comp.
        mapped_comp: The matched labels of the result masks. A list of length
            equal to the number of frames. Each element is a list with the
            matched labels of the result masks in the respective frame. The
            elements are in the same order as the corresponding elements in
            mapped_ref.

    Returns:
        The track fractions metric.
    """
    comp_fractions = calculate_fractions_fo_computed_tracks(
        ref_tracks, labels_ref, mapped_ref, mapped_comp)
    # Calculate the track fractions with respect to the reference tracks
    tfs = {k: 0 for k in ref_tracks[:, 0]}
    for _, v in sorted(comp_fractions.items()):
        for k2, v2 in sorted(v.items()):
            if tfs[k2] == 1:
                continue
            if v2 > tfs[k2]:
                tfs[k2] = v2
            # ###
            # The next lines do not make sense, because it causes instabilities.
            # Exactly same results with different order of labels can lead to
            # different metrics. They are kept, because they are in the original
            # codebase for the CTC. Remove the break to solve the issue.
            # Note: Does not make real differences in most cases.
            # ###
            if v2 == 1:
                break
    # Filter out undetected tracks. They should not be counted.
    tfs = [x for k, x in tfs.items() if x > 0]
    if len(tfs) == 0:
        return 0
    return np.mean(np.asarray(tfs))



import warnings
import numpy as np


def valid_parents(
        tracks: np.ndarray,
):
    """
    Checks if all parents are >= 0.

    Args:
        tracks: The result tracks. Numpy array (n x 4) the columns are:
            0: Label
            1: Birth
            2: End
            3: Parent

    Returns:
        1 if all parents are >= 0, 0 otherwise.

    """
    is_valid = 1
    for track in tracks:
        if track[3] < 0:
            warnings.warn(f"Invalid parent: {track}", UserWarning)
            is_valid = 0
    return int(is_valid)


def unique_labels(
        tracks: np.ndarray,
):
    """
    Checks if all labels are unique.

    Args:
        tracks: The result tracks. Numpy array (n x 4) the columns are:
            0: Label
            1: Birth
            2: End
            3: Parent

    Returns:
        1 if all labels are unique, 0 otherwise.

    """
    is_valid = 1
    labels = tracks[:, 0]
    if len(np.unique(labels)) != len(labels):
        warnings.warn("Labels are not unique.", UserWarning)
        is_valid = 0
    return int(is_valid)


def valid_parent_links(
        tracks: np.ndarray,
):
    """
    Checks if all parent links are valid.

    Args:
        tracks: The result tracks. Numpy array (n x 4) the columns are:
            0: Label
            1: Birth
            2: End
            3: Parent

    Returns:
        1 if all parent links are valid, 0 otherwise.

    """
    is_valid = 1
    for track in tracks:
        _, birth, _, parent = track
        if parent != 0:
            parent_idx = np.argwhere(tracks[:, 0] == parent).squeeze()
            assert parent_idx.size == 1, parent_idx
            parent_track = tracks[parent_idx]
            _, _, parent_end, _ = parent_track
            if parent_end >= birth:
                warnings.warn(
                    f"Parent ends after child starts: {track} {parent_track}",
                    UserWarning)
                is_valid = 0
    return int(is_valid)


def valid_ends(
        tracks: np.ndarray,
):
    """
    Checks if the end is not before the birth.

    Args:
        tracks: The result tracks. Numpy array (n x 4) with The columns are:
            0: Label
            1: Birth
            2: End
            3: Parent

    Returns:
        1 if all parent links are valid, 0 otherwise.

    """
    is_valid = 1
    for track in tracks:
        i, birth, end, parent = track
        if end < birth:
            warnings.warn(
                f"Track ends before birth: {i} {birth} {end} {parent}",
                UserWarning)
            is_valid = 0
    return int(is_valid)


def inspect_masks(
        frames: list,
        masks: list,
        labels_in_frames: list,
):
    """
    Inspects the masks for invalid labels.

    Args:
        frames: The frames to inspect.
        masks: The mask files to inspect.
        labels_in_frames: The present labels corresponding to the file in
            "masks".

    Returns:
        1 if all labels are valid, 0 otherwise.

    """
    is_valid = 1
    for labels_in_frame, file, frame in zip(labels_in_frames, masks, frames):
        for label in labels_in_frame:
            if label != 0:
                if label not in frame:
                    warnings.warn(
                        f"Label {label} in mask but not in res_track {file}.",
                        UserWarning)
                    is_valid = 0
        for label in frame:
            if label not in labels_in_frame:
                warnings.warn(
                    f"Label {label} in res_track but not in mask {file}.",
                    UserWarning)
                is_valid = 0
    return int(is_valid)


def no_empty_frames(
        frames: list,
):
    """
    Checks if there are empty frames.

    Args:
        frames: The frames to inspect.

    Returns:
        1 if there are no empty frames, 0 otherwise.

    """
    is_valid = 1
    for i, f in enumerate(frames):
        if len(f) == 0:
            warnings.warn(f"Empty frame {i}.", UserWarning)
            is_valid = 0
    return int(is_valid)


def no_empty_tracking_result(
        tracks: np.ndarray
):
    """
    Checks if there is at least one detection in th results.

    Args:
        tracks: The tracks to inspect

    Returns:
        1 if there are detections, 0 otherwise.
    """
    is_valid = 1
    if len(tracks) == 0:
        warnings.warn("No tracks in result.", UserWarning)
        is_valid = 0
    return is_valid


def valid(
        masks: list,
        tracks: np.ndarray,
        labels_in_frames: list,
):
    """
    Checks if the cell tracking result is valid. The result is valid if...
    - ...all parents are >= 0
    - ...all labels are unique
    - ...parents endet before children are born
    (- ...all labels are used)
    - ...all labels are in the frames they are used to be
    - ...frames are not empty

    Args:
        masks: The masks of the result.
        tracks: The result tracks.
        labels_in_frames: The present labels corresponding to the file in
            "masks".

    Returns:
        1 if the result is valid, 0 otherwise.

    """
    is_valid = 1
    # If tracks is empty, the result is invalid
    is_valid = no_empty_tracking_result(tracks)
    # Get the labels in each frame
    num_frames = max(tracks[:, 2].max() + 1, len(masks))
    frames = [[] for _ in range(num_frames)]
    for track in tracks:
        label, birth, end, _ = track
        for frame in range(birth, end + 1):
            frames[frame].append(label)
    # Check if all parents are >= 0
    is_valid *= valid_parents(tracks)
    # Check if all labels are unique
    is_valid *= unique_labels(tracks)
    # Check if parents ends before children are born
    is_valid *= valid_parent_links(tracks)
    # Check if end is not before birth
    is_valid *= valid_ends(tracks)
    # Check if all labels are in the frames they are used to be
    is_valid *= inspect_masks(frames, masks, labels_in_frames)
    # Check if frames are empty
    no_empty_frames(frames)  # Should this make the validation irregular?
    return int(is_valid)



import numpy as np


def seg(
        labels_ref: list,
        intersection_over_unions: list,
):
    """
    Calculates the segmentation metric. The metric describes the average overlap
    between the reference labels and an assigned result labels.

    Args:
        labels_ref: A list of lists with the labels of the ground truth masks
            according to the respective frame.
        intersection_over_unions: A list of lists with the intersection over
            union values of matched reference and computed result labels.

    Returns:
        The segmentation metric, the number of true positives and the number of
        false negatives.

    """
    number_of_reference_labels = np.sum([len(x) for x in labels_ref])
    intersection_over_unions = np.concatenate(intersection_over_unions)
    total_intersection = np.sum(intersection_over_unions)
    seg_measure = total_intersection / np.maximum(number_of_reference_labels, 1)
    return float(seg_measure)




def bio(
        ct: float,
        tf: float,
        bc: float,
        cca: float,
):
    """
    Computes the BIO. As described by
        [celltrackingchallenge](http://celltrackingchallenge.net/).
    It is the average of the CT, TF, BC, and CCA metrics. If a metric is not
    available, it is not considered in the average.

    Args:
        ct: The complete tracking metric.
        tf: The track fractions metric.
        bc: The branching correctness metric.
        cca: The cell cycle accuracy metric.

    Returns:
        The BIO metric.
    """
    total_metrics = 0
    if ct is not None:
        total_metrics += 1
    else:
        ct = 0
    if tf is not None:
        total_metrics += 1
    else:
        tf = 0
    if bc is not None:
        total_metrics += 1
    else:
        bc = 0
    if cca is not None:
        total_metrics += 1
    else:
        cca = 0
    bio_score = (ct + tf + bc + cca) / total_metrics
    return bio_score



import numpy as np

def faf(
        labels_comp: list,
        mapped_comp: list
):
    """
    Computes average number of false alarms per frame. As described by
    [motchallenge](https://motchallenge.net/).

    Args:
        labels_comp: The labels of the computed masks.
        mapped_comp: The matched labels of the result masks.

    Returns:
        The FAF metric.
    """

    fp = 0
    frames = len(labels_comp)

    for comp, m_comp in zip(labels_comp, mapped_comp):
        uniques, counts = np.unique(m_comp, return_counts=True)
        uniques = uniques[counts == 1]
        fp += len(comp) - len(uniques)

    faf_score = fp / frames

    res = {
        "FAF": faf_score,
        "Frames": frames,
    }

    return res



import numpy as np


def mota(
        labels_ref: list,
        labels_comp: list,
        mapped_ref: list,
        mapped_comp: list
):
    """
    Computes the MOTA metric. As described in the paper,
         "Evaluating Multiple Object Tracking Performance:
          The CLEAR MOT Metrics."
           - Keni Bernardin and Rainer Stiefelhagen, EURASIP 2008

    Args:
        labels_comp: The labels of the computed masks. A list of length equal
            to the number of frames. Each element is a list with the labels of
            the computed masks in the respective frame.
        labels_ref: The labels of the ground truth masks. A list of length
            equal to the number of frames. Each element is a list with the
            labels of the ground truth masks in the respective frame.
        mapped_ref: The matched labels of the ground truth masks. A list of
            length equal to the number of frames. Each element is a list with
            the matched labels of the ground truth masks in the respective
            frame. The elements are in the same order as the corresponding
            elements in mapped_comp.
        mapped_comp: The matched labels of the result masks. A list of length
            equal to the number of frames. Each element is a list with the
            matched labels of the result masks in the respective frame. The
            elements are in the same order as the corresponding elements in
            mapped_ref.

    Returns:
        The MOTA tracks metric.
    """
    tp, fp, fn, idsw, multi_assignments = 0, 0, 0, 0, 0

    max_label_gt = int(np.max(np.concatenate(labels_ref)))
    matches = np.zeros(max_label_gt + 1)
    for ref, comp, m_ref, m_comp in zip(
            labels_ref, labels_comp, mapped_ref, mapped_comp):
        # Calculate metrics
        _, counts = np.unique(m_comp, return_counts=True)
        tp += len(m_ref)
        fn += len(ref) - len(m_ref)
        fp += len(comp) - len(m_comp) + np.sum(counts[counts > 1] - 1)
        multi_assignments += np.sum(counts[counts > 1] - 1)
        idsw += np.sum((matches[m_ref] != m_comp) & (matches[m_ref] != 0))
        # Update the match cache
        matches[m_ref] = m_comp

    mota_score = 1 - (fn + fp + idsw + multi_assignments) / (tp + fn)
    precision = tp / (tp + fp)
    recall = tp / (tp + fn)

    res = {
        "MOTA": mota_score,
        "TP": tp,
        "FP": fp,
        "FN": fn,
        "IDSW": idsw,
        "MULTI-ASSIGNMENTS": multi_assignments,
        "Precision": precision,
        "Recall": recall
    }
    return res
